@@ -613,6 +613,19 @@ void pqKeyFrameEditor::writeKeyFrameData()
 
   this->Internal->Cue->blockSignals(prev);
   this->Internal->Cue->triggerKeyFramesModified();
+
+  // Reload current timestep, force deleting outdated geometry cache
+  pqSMAdaptor::setElementProperty(
+    this->Internal->Scene->getProxy()->GetProperty("ForceDisableCaching", 1), 1);
+  this->Internal->Scene->getProxy()->UpdateProperty("ForceDisableCaching", 1);
+  pqSMAdaptor::setElementProperty(this->Internal->Scene->getProxy()->GetProperty("AnimationTime"),
+    this->Internal->Scene->getAnimationTime());
+  this->Internal->Scene->getProxy()->UpdateProperty("AnimationTime", 1);
+  this->Internal->Scene->getProxy()->UpdateVTKObjects();
+  pqSMAdaptor::setElementProperty(
+    this->Internal->Scene->getProxy()->GetProperty("ForceDisableCaching", 1), 0);
+  this->Internal->Scene->getProxy()->UpdateProperty("ForceDisableCaching", 1);
+
   END_UNDO_SET();
 }
 
